@@ -2,7 +2,7 @@
 #include "nameser.h"
 
 
-static PyTypeObject DNSHostResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject DNSHostResultType;
 
 static PyStructSequence_Field dns_host_result_fields[] = {
     {"name", ""},
@@ -18,7 +18,7 @@ static PyStructSequence_Desc dns_host_result_desc = {
     3
 };
 
-static PyTypeObject DNSNameinfoResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject DNSNameinfoResultType;
 
 static PyStructSequence_Field dns_nameinfo_result_fields[] = {
     {"node", ""},
@@ -33,7 +33,7 @@ static PyStructSequence_Desc dns_nameinfo_result_desc = {
     2
 };
 
-static PyTypeObject DNSAddrinfoResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject DNSAddrinfoResultType;
 
 static PyStructSequence_Field dns_addrinfo_result_fields[] = {
     {"family", ""},
@@ -51,7 +51,7 @@ static PyStructSequence_Desc dns_addrinfo_result_desc = {
     5
 };
 
-static PyTypeObject DNSQueryMXResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject DNSQueryMXResultType;
 
 static PyStructSequence_Field dns_query_mx_result_fields[] = {
     {"host", ""},
@@ -66,7 +66,7 @@ static PyStructSequence_Desc dns_query_mx_result_desc = {
     2
 };
 
-static PyTypeObject DNSQuerySRVResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject DNSQuerySRVResultType;
 
 static PyStructSequence_Field dns_query_srv_result_fields[] = {
     {"host", ""},
@@ -83,7 +83,7 @@ static PyStructSequence_Desc dns_query_srv_result_desc = {
     4
 };
 
-static PyTypeObject DNSQueryNAPTRResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject DNSQueryNAPTRResultType;
 
 static PyStructSequence_Field dns_query_naptr_result_fields[] = {
     {"order", ""},
@@ -1184,7 +1184,7 @@ DNSResolver_func_getaddrinfo(DNSResolver *self, PyObject *args, PyObject *kwargs
 
     r = uv_getaddrinfo(UV_LOOP(self), handle, &getaddrinfo_cb, name, port_str, &hints);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_DNSError);
+        RAISE_UV_EXCEPTION(UV_LOOP(self), PyExc_DNSError);
         goto error;
     }
 
@@ -1477,7 +1477,6 @@ DNSResolver_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 DNSResolver_tp_traverse(DNSResolver *self, visitproc visit, void *arg)
 {
-    Py_VISIT(self->data);
     Py_VISIT(self->loop);
     return 0;
 }
@@ -1486,7 +1485,6 @@ DNSResolver_tp_traverse(DNSResolver *self, visitproc visit, void *arg)
 static int
 DNSResolver_tp_clear(DNSResolver *self)
 {
-    Py_CLEAR(self->data);
     Py_CLEAR(self->loop);
     return 0;
 }
@@ -1518,7 +1516,6 @@ DNSResolver_tp_methods[] = {
 
 static PyMemberDef DNSResolver_tp_members[] = {
     {"loop", T_OBJECT_EX, offsetof(DNSResolver, loop), READONLY, "Loop where this DNSResolver is running on."},
-    {"data", T_OBJECT, offsetof(DNSResolver, data), 0, "Arbitrary data."},
     {NULL}
 };
 
