@@ -1,5 +1,5 @@
 
-static PyTypeObject StatResultType = {0, 0, 0, 0, 0, 0};
+static PyTypeObject StatResultType;
 
 static PyStructSequence_Field stat_result_fields[] = {
     {"st_mode",        "protection bits"},
@@ -61,8 +61,7 @@ format_time(time_t sec, unsigned long nsec)
 static void
 process_stat(uv_fs_t* req, PyObject **path, PyObject **stat_data, PyObject **errorno) {
     struct stat *st;
-    unsigned long ansec, mnsec, cnsec, bsec, bnsec;
-    PyObject *tmp;
+    unsigned long ansec, mnsec, cnsec, bnsec;
 
     ASSERT(req);
     ASSERT(req->fs_type == UV_FS_STAT || req->fs_type == UV_FS_LSTAT || req->fs_type == UV_FS_FSTAT);
@@ -165,14 +164,12 @@ process_stat(uv_fs_t* req, PyObject **path, PyObject **stat_data, PyObject **err
     PyStructSequence_SET_ITEM(*stat_data, 14, Py_None);
 #endif
 #ifdef HAVE_STRUCT_STAT_ST_BIRTHTIME
-    bsec = (long)st->st_birthtime;
 #ifdef HAVE_STAT_TV_NSEC2
     bnsec = st->st_birthtimespec.tv_nsec;
 #else
     bnsec = 0;
 #endif
-    tmp = PyInt_FromLong((long)bsec);
-    PyStructSequence_SET_ITEM(*stat_data, 15, tmp);
+    PyStructSequence_SET_ITEM(*stat_data, 15, format_time(st->st_birthtime, bnsec));
 #else
     Py_INCREF(Py_None);
     PyStructSequence_SET_ITEM(*stat_data, 15, Py_None);
@@ -620,7 +617,11 @@ process_open(uv_fs_t* req, PyObject **path, PyObject **fd, PyObject **errorno)
     } else {
         *errorno = Py_None;
         Py_INCREF(Py_None);
+#ifdef PYUV_WINDOWS
+        *fd = PyInt_FromLong((long)PYUV_FD_TO_WIN32_HANDLE(req->result));
+#else
         *fd = PyInt_FromLong((long)req->result);
+#endif
     }
 }
 
@@ -1112,7 +1113,7 @@ stat_func(PyObject *args, PyObject *kwargs, int type)
         r = uv_fs_lstat(loop->uv_loop, fs_req, path, (callback != NULL) ? stat_cb : NULL);
     }
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -1179,6 +1180,14 @@ FS_func_fstat(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -1202,7 +1211,7 @@ FS_func_fstat(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_fstat(loop->uv_loop, fs_req, fd, (callback != NULL) ? stat_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -1275,7 +1284,7 @@ FS_func_unlink(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_unlink(loop->uv_loop, fs_req, path, (callback != NULL) ? unlink_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1339,7 +1348,7 @@ FS_func_mkdir(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_mkdir(loop->uv_loop, fs_req, path, mode, (callback != NULL) ? mkdir_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1403,7 +1412,7 @@ FS_func_rmdir(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_rmdir(loop->uv_loop, fs_req, path, (callback != NULL) ? rmdir_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1467,7 +1476,7 @@ FS_func_rename(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_rename(loop->uv_loop, fs_req, path, new_path, (callback != NULL) ? rename_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1531,7 +1540,7 @@ FS_func_chmod(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_chmod(loop->uv_loop, fs_req, path, mode, (callback != NULL) ? chmod_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1573,6 +1582,14 @@ FS_func_fchmod(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -1594,7 +1611,7 @@ FS_func_fchmod(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_fchmod(loop->uv_loop, fs_req, fd, mode, (callback != NULL) ? chmod_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1658,7 +1675,7 @@ FS_func_link(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_link(loop->uv_loop, fs_req, path, new_path, (callback != NULL) ? link_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1722,7 +1739,7 @@ FS_func_symlink(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_symlink(loop->uv_loop, fs_req, path, new_path, flags, (callback != NULL) ? symlink_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1788,7 +1805,7 @@ FS_func_readlink(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_readlink(loop->uv_loop, fs_req, path, (callback != NULL) ? readlink_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -1860,7 +1877,7 @@ FS_func_chown(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_chown(loop->uv_loop, fs_req, path, uid, gid, (callback != NULL) ? chown_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1902,6 +1919,14 @@ FS_func_fchown(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -1923,7 +1948,7 @@ FS_func_fchown(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_fchown(loop->uv_loop, fs_req, fd, uid, gid, (callback != NULL) ? chown_cb : NULL);
     if (r != 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -1989,7 +2014,7 @@ FS_func_open(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_open(loop->uv_loop, fs_req, path, flags, mode, (callback != NULL) ? open_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -2040,6 +2065,14 @@ FS_func_close(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2061,7 +2094,7 @@ FS_func_close(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_close(loop->uv_loop, fs_req, fd, (callback != NULL) ? close_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -2104,6 +2137,14 @@ FS_func_read(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2136,7 +2177,7 @@ FS_func_read(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_read(loop->uv_loop, fs_req, fd, req_data->buf.base, length, offset, (callback != NULL) ? read_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -2195,6 +2236,14 @@ FS_func_write(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2229,7 +2278,7 @@ FS_func_write(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_write(loop->uv_loop, fs_req, fd, req_data->buf.base, req_data->buf.len, offset, (callback != NULL) ? write_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -2285,6 +2334,14 @@ FS_func_fsync(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2306,7 +2363,7 @@ FS_func_fsync(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_fsync(loop->uv_loop, fs_req, fd, (callback != NULL) ? fsync_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -2348,6 +2405,14 @@ FS_func_fdatasync(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2369,7 +2434,7 @@ FS_func_fdatasync(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_fdatasync(loop->uv_loop, fs_req, fd, (callback != NULL) ? fsync_cb : NULL);
     if (r != 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -2411,6 +2476,14 @@ FS_func_ftruncate(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2432,7 +2505,7 @@ FS_func_ftruncate(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_ftruncate(loop->uv_loop, fs_req, fd, offset, (callback != NULL) ? ftruncate_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -2499,7 +2572,7 @@ FS_func_readdir(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_readdir(loop->uv_loop, fs_req, path, flags, (callback != NULL) ? readdir_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -2550,6 +2623,19 @@ FS_func_sendfile(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    in_fd = PYUV_WIN32_HANDLE_TO_FD(in_fd);
+    if (in_fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+    out_fd = PYUV_WIN32_HANDLE_TO_FD(out_fd);
+    if (out_fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2573,7 +2659,7 @@ FS_func_sendfile(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_sendfile(loop->uv_loop, fs_req, out_fd, in_fd, in_offset, length, (callback != NULL) ? sendfile_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         ret = NULL;
         goto end;
     }
@@ -2647,7 +2733,7 @@ FS_func_utime(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_utime(loop->uv_loop, fs_req, path, atime, mtime, (callback != NULL) ? utime_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -2690,6 +2776,14 @@ FS_func_futime(PyObject *obj, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
+#ifdef PYUV_WINDOWS
+    fd = PYUV_WIN32_HANDLE_TO_FD(fd);
+    if (fd == -1) {
+        PyErr_SetFromWindowsErr(0);
+        return NULL;
+    }
+#endif
+
     fs_req = PyMem_Malloc(sizeof(uv_fs_t));
     if (!fs_req) {
         PyErr_NoMemory();
@@ -2711,7 +2805,7 @@ FS_func_futime(PyObject *obj, PyObject *args, PyObject *kwargs)
     fs_req->data = (void *)req_data;
     r = uv_fs_futime(loop->uv_loop, fs_req, fd, atime, mtime, (callback != NULL) ? utime_cb : NULL);
     if (r < 0) {
-        raise_uv_exception(loop, PyExc_FSError);
+        RAISE_UV_EXCEPTION(loop->uv_loop, PyExc_FSError);
         goto end;
     }
 
@@ -2764,46 +2858,6 @@ FS_methods[] = {
 static PyObject* PyExc_FSEventError;
 
 
-static void
-on_fsevent_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    FSEvent *self;
-    PyObject *result;
-
-    ASSERT(handle);
-    self = (FSEvent *)handle->data;
-    ASSERT(self);
-
-    if (self->on_close_cb != Py_None) {
-        result = PyObject_CallFunctionObjArgs(self->on_close_cb, self, NULL);
-        if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_close_cb);
-        }
-        Py_XDECREF(result);
-    }
-
-    handle->data = NULL;
-    PyMem_Free(handle);
-
-    /* Refcount was increased in func_close */
-    Py_DECREF(self);
-
-    PyGILState_Release(gstate);
-}
-
-
-static void
-on_fsevent_dealloc_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    ASSERT(handle);
-    handle->data = NULL;
-    PyMem_Free(handle);
-    PyGILState_Release(gstate);
-}
-
-
 static void
 on_fsevent_callback(uv_fs_event_t *handle, const char *filename, int events, int status)
 {
@@ -2827,7 +2881,7 @@ on_fsevent_callback(uv_fs_event_t *handle, const char *filename, int events, int
     }
 
     if (status < 0) {
-        uv_err_t err = uv_last_error(UV_LOOP(self));
+        uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
         errorno = PyInt_FromLong((long)err.code);
     } else {
         errorno = Py_None;
@@ -2862,7 +2916,7 @@ FSEvent_func_start(FSEvent *self, PyObject *args, PyObject *kwargs)
 
     tmp = NULL;
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_FSEventError, "FSEvent was already started");
         return NULL;
     }
@@ -2882,13 +2936,13 @@ FSEvent_func_start(FSEvent *self, PyObject *args, PyObject *kwargs)
         return NULL;
     }
     fs_event->data = (void *)self;
-    self->uv_handle = fs_event;
+    UV_HANDLE(self) = (uv_handle_t *)fs_event;
 
-    r = uv_fs_event_init(UV_LOOP(self), self->uv_handle, path, on_fsevent_callback, flags);
+    r = uv_fs_event_init(UV_HANDLE_LOOP(self), fs_event, path, on_fsevent_callback, flags);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_FSEventError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_FSEventError);
         PyMem_Free(fs_event);
-        self->uv_handle = NULL;
+        UV_HANDLE(self) = NULL;
         return NULL;
     }
 
@@ -2902,34 +2956,14 @@ FSEvent_func_start(FSEvent *self, PyObject *args, PyObject *kwargs)
 
 
 static PyObject *
-FSEvent_func_close(FSEvent *self, PyObject *args)
+FSEvent_filename_get(FSEvent *self, void *closure)
 {
-    PyObject *callback = Py_None;
+    UNUSED_ARG(closure);
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_FSEventError, "FSEvent has not been started");
-        return NULL;
+    if (!UV_HANDLE(self)) {
+        Py_RETURN_NONE;
     }
-
-    if (!PyArg_ParseTuple(args, "|O:close", &callback)) {
-        return NULL;
-    }
-
-    if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    Py_INCREF(callback);
-    self->on_close_cb = callback;
-
-    /* Increase refcount so that object is not removed before the callback is called */
-    Py_INCREF(self);
-
-    uv_close((uv_handle_t *)self->uv_handle, on_fsevent_close);
-    self->uv_handle = NULL;
-
-    Py_RETURN_NONE;
+    return PYUVString_FromString(((uv_fs_event_t *)UV_HANDLE(self))->filename);
 }
 
 
@@ -2941,7 +2975,7 @@ FSEvent_tp_init(FSEvent *self, PyObject *args, PyObject *kwargs)
 
     UNUSED_ARG(kwargs);
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_FSEventError, "Object already initialized");
         return -1;
     }
@@ -2950,9 +2984,9 @@ FSEvent_tp_init(FSEvent *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    tmp = (PyObject *)self->loop;
+    tmp = (PyObject *)((Handle *)self)->loop;
     Py_INCREF(loop);
-    self->loop = loop;
+    ((Handle *)self)->loop = loop;
     Py_XDECREF(tmp);
 
     return 0;
@@ -2962,11 +2996,10 @@ FSEvent_tp_init(FSEvent *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 FSEvent_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    FSEvent *self = (FSEvent *)PyType_GenericNew(type, args, kwargs);
+    FSEvent *self = (FSEvent *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
         return NULL;
     }
-    self->uv_handle = NULL;
     return (PyObject *)self;
 }
 
@@ -2974,10 +3007,8 @@ FSEvent_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 FSEvent_tp_traverse(FSEvent *self, visitproc visit, void *arg)
 {
-    Py_VISIT(self->data);
     Py_VISIT(self->on_fsevent_cb);
-    Py_VISIT(self->on_close_cb);
-    Py_VISIT(self->loop);
+    HandleType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
 
@@ -2985,37 +3016,21 @@ FSEvent_tp_traverse(FSEvent *self, visitproc visit, void *arg)
 static int
 FSEvent_tp_clear(FSEvent *self)
 {
-    Py_CLEAR(self->data);
     Py_CLEAR(self->on_fsevent_cb);
-    Py_CLEAR(self->on_close_cb);
-    Py_CLEAR(self->loop);
+    HandleType.tp_clear((PyObject *)self);
     return 0;
 }
 
 
-static void
-FSEvent_tp_dealloc(FSEvent *self)
-{
-    if (self->uv_handle) {
-        uv_close((uv_handle_t *)self->uv_handle, on_fsevent_dealloc_close);
-        self->uv_handle = NULL;
-    }
-    FSEvent_tp_clear(self);
-    Py_TYPE(self)->tp_free((PyObject *)self);
-}
-
-
 static PyMethodDef
 FSEvent_tp_methods[] = {
     { "start", (PyCFunction)FSEvent_func_start, METH_VARARGS|METH_KEYWORDS, "Start the FSEvent." },
-    { "close", (PyCFunction)FSEvent_func_close, METH_VARARGS, "Close the FSEvent." },
     { NULL }
 };
 
 
-static PyMemberDef FSEvent_tp_members[] = {
-    {"loop", T_OBJECT_EX, offsetof(FSEvent, loop), READONLY, "Loop where this FSEvent is running on."},
-    {"data", T_OBJECT, offsetof(FSEvent, data), 0, "Arbitrary data."},
+static PyGetSetDef FSEvent_tp_getsets[] = {
+    {"filename", (getter)FSEvent_filename_get, NULL, "Name of the file/directory being monitored.", NULL},
     {NULL}
 };
 
@@ -3025,7 +3040,7 @@ static PyTypeObject FSEventType = {
     "pyuv.fs.FSEvent",                                              /*tp_name*/
     sizeof(FSEvent),                                                /*tp_basicsize*/
     0,                                                              /*tp_itemsize*/
-    (destructor)FSEvent_tp_dealloc,                                 /*tp_dealloc*/
+    0,                                                              /*tp_dealloc*/
     0,                                                              /*tp_print*/
     0,                                                              /*tp_getattr*/
     0,                                                              /*tp_setattr*/
@@ -3049,8 +3064,8 @@ static PyTypeObject FSEventType = {
     0,                                                              /*tp_iter*/
     0,                                                              /*tp_iternext*/
     FSEvent_tp_methods,                                             /*tp_methods*/
-    FSEvent_tp_members,                                             /*tp_members*/
-    0,                                                              /*tp_getsets*/
+    0,                                                              /*tp_members*/
+    FSEvent_tp_getsets,                                             /*tp_getsets*/
     0,                                                              /*tp_base*/
     0,                                                              /*tp_dict*/
     0,                                                              /*tp_descr_get*/
@@ -3090,6 +3105,7 @@ init_fs(void)
     PyModule_AddIntMacro(module, UV_FS_EVENT_WATCH_ENTRY);
     PyModule_AddIntMacro(module, UV_FS_EVENT_STAT);
 
+    FSEventType.tp_base = &HandleType;
     PyUVModule_AddType(module, "FSEvent", &FSEventType);
 
     return module;
