@@ -7,7 +7,6 @@ on_pipe_connection(uv_stream_t* server, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     Pipe *self;
-    IOStream *base;
     PyObject *result, *py_errorno;
     ASSERT(server);
 
@@ -16,10 +15,8 @@ on_pipe_connection(uv_stream_t* server, int status)
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
 
-    base = (IOStream *)self;
-
     if (status != 0) {
-        uv_err_t err = uv_last_error(UV_LOOP(base));
+        uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
         py_errorno = PyInt_FromLong((long)err.code);
     } else {
         py_errorno = Py_None;
@@ -42,24 +39,21 @@ static void
 on_pipe_client_connection(uv_connect_t *req, int status)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    iostream_req_data_t* req_data;
+    stream_req_data_t* req_data;
     Pipe *self;
-    IOStream *base;
     PyObject *callback, *result, *py_errorno;
     ASSERT(req);
 
-    req_data = (iostream_req_data_t *)req->data;
-    self = (Pipe *)req_data->obj;
+    req_data = (stream_req_data_t *)req->data;
+    self = (Pipe *)req->handle->data;
     callback = req_data->callback;
 
     ASSERT(self);
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
 
-    base = (IOStream *)self;
-
     if (status != 0) {
-        uv_err_t err = uv_last_error(UV_LOOP(base));
+        uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
         py_errorno = PyInt_FromLong(err.code);
     } else {
         py_errorno = Py_None;
@@ -87,11 +81,11 @@ on_pipe_read2(uv_pipe_t* handle, int nread, uv_buf_t buf, uv_handle_type pending
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
     uv_err_t err;
-    IOStream *self;
+    Stream *self;
     PyObject *result, *data, *py_errorno, *py_pending;
     ASSERT(handle);
 
-    self = (IOStream *)handle->data;
+    self = (Stream *)handle->data;
     ASSERT(self);
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
@@ -105,7 +99,7 @@ on_pipe_read2(uv_pipe_t* handle, int nread, uv_buf_t buf, uv_handle_type pending
     } else if (nread < 0) {
         data = Py_None;
         Py_INCREF(Py_None);
-        err = uv_last_error(UV_LOOP(self));
+        err = uv_last_error(UV_HANDLE_LOOP(self));
         py_errorno = PyInt_FromLong((long)err.code);
     }
 
@@ -131,20 +125,15 @@ Pipe_func_bind(Pipe *self, PyObject *args)
     int r;
     char *name;
 
-    IOStream *base = (IOStream *)self;
-
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "already closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "s:bind", &name)) {
         return NULL;
     }
 
-    r = uv_pipe_bind((uv_pipe_t *)base->uv_handle, name);
+    r = uv_pipe_bind((uv_pipe_t *)UV_HANDLE(self), name);
     if (r != 0) {
-        raise_uv_exception(base->loop, PyExc_PipeError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
         return NULL;
     }
 
@@ -158,15 +147,10 @@ Pipe_func_listen(Pipe *self, PyObject *args)
     int r, backlog;
     PyObject *callback, *tmp;
 
-    IOStream *base = (IOStream *)self;
-
     backlog = 128;
     tmp = NULL;
 
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "already closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O|i:listen", &callback, &backlog)) {
         return NULL;
@@ -182,9 +166,9 @@ Pipe_func_listen(Pipe *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_listen(base->uv_handle, backlog, on_pipe_connection);
+    r = uv_listen((uv_stream_t *)UV_HANDLE(self), backlog, on_pipe_connection);
     if (r != 0) {
-        raise_uv_exception(base->loop, PyExc_PipeError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
         return NULL;
     }
 
@@ -203,25 +187,20 @@ Pipe_func_accept(Pipe *self, PyObject *args)
     int r;
     PyObject *client;
 
-    IOStream *base = (IOStream *)self;
-
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "already closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:accept", &client)) {
         return NULL;
     }
 
-    if (!PyObject_IsSubclass((PyObject *)client->ob_type, (PyObject *)&IOStreamType)) {
+    if (!PyObject_IsSubclass((PyObject *)client->ob_type, (PyObject *)&StreamType)) {
         PyErr_SetString(PyExc_TypeError, "Only stream objects are supported for accept");
         return NULL;
     }
 
-    r = uv_accept(base->uv_handle, ((IOStream *)client)->uv_handle);
+    r = uv_accept((uv_stream_t *)UV_HANDLE(self), (uv_stream_t *)UV_HANDLE(client));
     if (r != 0) {
-        raise_uv_exception(base->loop, PyExc_PipeError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
         return NULL;
     }
 
@@ -234,15 +213,10 @@ Pipe_func_connect(Pipe *self, PyObject *args)
 {
     char *name;
     uv_connect_t *connect_req = NULL;
-    iostream_req_data_t *req_data = NULL;
+    stream_req_data_t *req_data = NULL;
     PyObject *callback;
 
-    IOStream *base = (IOStream *)self;
-
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "sO:connect", &name, &callback)) {
         return NULL;
@@ -259,18 +233,17 @@ Pipe_func_connect(Pipe *self, PyObject *args)
         goto error;
     }
 
-    req_data = (iostream_req_data_t*) PyMem_Malloc(sizeof(iostream_req_data_t));
+    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
     if (!req_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data->obj = (PyObject *)self;
     Py_INCREF(callback);
     req_data->callback = callback;
     connect_req->data = (void *)req_data;
 
-    uv_pipe_connect(connect_req, (uv_pipe_t *)base->uv_handle, name, on_pipe_client_connection);
+    uv_pipe_connect(connect_req, (uv_pipe_t *)UV_HANDLE(self), name, on_pipe_client_connection);
 
     Py_RETURN_NONE;
 
@@ -291,18 +264,13 @@ Pipe_func_open(Pipe *self, PyObject *args)
 {
     int fd;
 
-    IOStream *base = (IOStream *)self;
-
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:open", &fd)) {
         return NULL;
     }
 
-    uv_pipe_open((uv_pipe_t *)base->uv_handle, fd);
+    uv_pipe_open((uv_pipe_t *)UV_HANDLE(self), fd);
 
     Py_RETURN_NONE;
 }
@@ -314,18 +282,13 @@ Pipe_func_pending_instances(Pipe *self, PyObject *args)
     /* This function applies to Windows only */
     int count;
 
-    IOStream *base = (IOStream *)self;
-
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:pending_instances", &count)) {
         return NULL;
     }
 
-    uv_pipe_pending_instances((uv_pipe_t *)base->uv_handle, count);
+    uv_pipe_pending_instances((uv_pipe_t *)UV_HANDLE(self), count);
 
     Py_RETURN_NONE;
 }
@@ -337,14 +300,9 @@ Pipe_func_start_read2(Pipe *self, PyObject *args)
     int r;
     PyObject *tmp, *callback;
 
-    IOStream *base = (IOStream *)self;
-
     tmp = NULL;
 
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "Pipe is closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:start_read2", &callback)) {
         return NULL;
@@ -355,15 +313,15 @@ Pipe_func_start_read2(Pipe *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_read2_start((uv_stream_t *)base->uv_handle, (uv_alloc_cb)on_iostream_alloc, (uv_read2_cb)on_pipe_read2);
+    r = uv_read2_start((uv_stream_t *)UV_HANDLE(self), (uv_alloc_cb)on_stream_alloc, (uv_read2_cb)on_pipe_read2);
     if (r != 0) {
-        raise_uv_exception(base->loop, PyExc_PipeError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
         return NULL;
     }
 
-    tmp = base->on_read_cb;
+    tmp = ((Stream *)self)->on_read_cb;
     Py_INCREF(callback);
-    base->on_read_cb = callback;
+    ((Stream *)self)->on_read_cb = callback;
     Py_XDECREF(tmp);
 
     Py_RETURN_NONE;
@@ -374,25 +332,17 @@ static PyObject *
 Pipe_func_write2(Pipe *self, PyObject *args)
 {
     int i, r, buf_count;
-    char *data_str, *tmp;
     Py_buffer pbuf;
-    Py_ssize_t data_len;
     PyObject *callback, *send_handle;
-    uv_buf_t tmpbuf;
     uv_buf_t *bufs = NULL;
     uv_write_t *wr = NULL;
-    iostream_req_data_t *req_data = NULL;
-    iostream_write_data_t *write_data = NULL;
-
-    IOStream *base = (IOStream *)self;
+    stream_req_data_t *req_data = NULL;
+    stream_write_data_t *write_data = NULL;
 
     buf_count = 0;
     callback = Py_None;
 
-    if (!base->uv_handle) {
-        PyErr_SetString(PyExc_PipeError, "Pipe is closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "s*O|O:write2", &pbuf, &send_handle, &callback)) {
         return NULL;
@@ -415,19 +365,18 @@ Pipe_func_write2(Pipe *self, PyObject *args)
         goto error;
     }
 
-    req_data = (iostream_req_data_t*) PyMem_Malloc(sizeof(iostream_req_data_t));
+    req_data = (stream_req_data_t*) PyMem_Malloc(sizeof(stream_req_data_t));
     if (!req_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    write_data = (iostream_write_data_t *) PyMem_Malloc(sizeof(iostream_write_data_t));
+    write_data = (stream_write_data_t *) PyMem_Malloc(sizeof(stream_write_data_t));
     if (!write_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data->obj = (PyObject *)base;
     Py_INCREF(callback);
     req_data->callback = callback;
     wr->data = (void *)req_data;
@@ -438,30 +387,20 @@ Pipe_func_write2(Pipe *self, PyObject *args)
         goto error;
     }
 
-    data_str = pbuf.buf;
-    data_len = pbuf.len;
-    tmp = (char *) PyMem_Malloc(data_len);
-    if (!tmp) {
-        PyErr_NoMemory();
-        goto error;
-    }
-    memcpy(tmp, data_str, data_len);
-    tmpbuf = uv_buf_init(tmp, data_len);
-    bufs[0] = tmpbuf;
+    bufs[0] = uv_buf_init(pbuf.buf, pbuf.len);
     buf_count = 1;
 
     write_data->bufs = bufs;
     write_data->buf_count = buf_count;
-    req_data->data = (void *)write_data;
+    write_data->view = pbuf;
+    req_data->data = write_data;
 
-    r = uv_write2(wr, base->uv_handle, bufs, buf_count, (uv_stream_t *)((IOStream *)send_handle)->uv_handle, on_iostream_write);
+    r = uv_write2(wr, (uv_stream_t *)UV_HANDLE(self), bufs, buf_count, (uv_stream_t *)UV_HANDLE(send_handle), on_stream_write);
     if (r != 0) {
-        raise_uv_exception(base->loop, PyExc_PipeError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
         goto error;
     }
 
-    PyBuffer_Release(&pbuf);
-
     Py_RETURN_NONE;
 
 error:
@@ -490,16 +429,14 @@ static int
 Pipe_tp_init(Pipe *self, PyObject *args, PyObject *kwargs)
 {
     int r;
-    uv_pipe_t *uv_stream;
+    uv_pipe_t *uv_pipe;
     Loop *loop;
     PyObject *tmp = NULL;
     PyObject *ipc = Py_False;
 
-    IOStream *base = (IOStream *)self;
-
     UNUSED_ARG(kwargs);
 
-    if (base->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_PipeError, "Object already initialized");
         return -1;
     }
@@ -508,26 +445,26 @@ Pipe_tp_init(Pipe *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    tmp = (PyObject *)base->loop;
+    tmp = (PyObject *)((Handle *)self)->loop;
     Py_INCREF(loop);
-    base->loop = loop;
+    ((Handle *)self)->loop = loop;
     Py_XDECREF(tmp);
 
-    uv_stream = PyMem_Malloc(sizeof(uv_pipe_t));
-    if (!uv_stream) {
+    uv_pipe = PyMem_Malloc(sizeof(uv_pipe_t));
+    if (!uv_pipe) {
         PyErr_NoMemory();
         Py_DECREF(loop);
         return -1;
     }
 
-    r = uv_pipe_init(UV_LOOP(base), (uv_pipe_t *)uv_stream, (ipc == Py_True) ? 1 : 0);
+    r = uv_pipe_init(UV_HANDLE_LOOP(self), uv_pipe, (ipc == Py_True) ? 1 : 0);
     if (r != 0) {
-        raise_uv_exception(base->loop, PyExc_PipeError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PipeError);
         Py_DECREF(loop);
         return -1;
     }
-    uv_stream->data = (void *)self;
-    base->uv_handle = (uv_stream_t *)uv_stream;
+    uv_pipe->data = (void *)self;
+    UV_HANDLE(self) = (uv_handle_t *)uv_pipe;
 
     return 0;
 }
@@ -536,7 +473,7 @@ Pipe_tp_init(Pipe *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Pipe_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Pipe *self = (Pipe *)IOStreamType.tp_new(type, args, kwargs);
+    Pipe *self = (Pipe *)StreamType.tp_new(type, args, kwargs);
     if (!self) {
         return NULL;
     }
@@ -548,7 +485,7 @@ static int
 Pipe_tp_traverse(Pipe *self, visitproc visit, void *arg)
 {
     Py_VISIT(self->on_new_connection_cb);
-    IOStreamType.tp_traverse((PyObject *)self, visit, arg);
+    StreamType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
 
@@ -557,7 +494,7 @@ static int
 Pipe_tp_clear(Pipe *self)
 {
     Py_CLEAR(self->on_new_connection_cb);
-    IOStreamType.tp_clear((PyObject *)self);
+    StreamType.tp_clear((PyObject *)self);
     return 0;
 }
 
