@@ -2,47 +2,6 @@
 static PyObject* PyExc_PrepareError;
 
 
-static void
-on_prepare_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    Prepare *self;
-    PyObject *result;
-
-    ASSERT(handle);
-
-    self = (Prepare *)handle->data;
-    ASSERT(self);
-
-    if (self->on_close_cb != Py_None) {
-        result = PyObject_CallFunctionObjArgs(self->on_close_cb, self, NULL);
-        if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_close_cb);
-        }
-        Py_XDECREF(result);
-    }
-
-    handle->data = NULL;
-    PyMem_Free(handle);
-
-    /* Refcount was increased in func_close */
-    Py_DECREF(self);
-
-    PyGILState_Release(gstate);
-}
-
-
-static void
-on_prepare_dealloc_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    ASSERT(handle);
-    handle->data = NULL;
-    PyMem_Free(handle);
-    PyGILState_Release(gstate);
-}
-
-
 static void
 on_prepare_callback(uv_prepare_t *handle, int status)
 {
@@ -77,10 +36,7 @@ Prepare_func_start(Prepare *self, PyObject *args)
 
     tmp = NULL;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_PrepareError, "Prepare is closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:start", &callback)) {
         return NULL;
@@ -91,9 +47,9 @@ Prepare_func_start(Prepare *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_prepare_start(self->uv_handle, on_prepare_callback);
+    r = uv_prepare_start((uv_prepare_t *)UV_HANDLE(self), on_prepare_callback);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_PrepareError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PrepareError);
         return NULL;
     }
 
@@ -111,61 +67,21 @@ Prepare_func_stop(Prepare *self)
 {
     int r;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_PrepareError, "Prepare is already closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_prepare_stop(self->uv_handle);
+    r = uv_prepare_stop((uv_prepare_t *)UV_HANDLE(self));
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_PrepareError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PrepareError);
         return NULL;
     }
 
+    Py_XDECREF(self->callback);
+    self->callback = NULL;
+
     Py_RETURN_NONE;
 }
 
 
-static PyObject *
-Prepare_func_close(Prepare *self, PyObject *args)
-{
-    PyObject *callback = Py_None;
-
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_PrepareError, "Prepare is already closed");
-        return NULL;
-    }
-
-    if (!PyArg_ParseTuple(args, "|O:close", &callback)) {
-        return NULL;
-    }
-
-    if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    Py_INCREF(callback);
-    self->on_close_cb = callback;
-
-    /* Increase refcount so that object is not removed before the callback is called */
-    Py_INCREF(self);
-
-    uv_close((uv_handle_t *)self->uv_handle, on_prepare_close);
-    self->uv_handle = NULL;
-
-    Py_RETURN_NONE;
-}
-
-
-static PyObject *
-Prepare_active_get(Prepare *self, void *closure)
-{
-    UNUSED_ARG(closure);
-    return PyBool_FromLong((long)uv_is_active((uv_handle_t *)self->uv_handle));
-}
-
-
 static int
 Prepare_tp_init(Prepare *self, PyObject *args, PyObject *kwargs)
 {
@@ -176,7 +92,7 @@ Prepare_tp_init(Prepare *self, PyObject *args, PyObject *kwargs)
 
     UNUSED_ARG(kwargs);
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_PrepareError, "Object already initialized");
         return -1;
     }
@@ -185,9 +101,9 @@ Prepare_tp_init(Prepare *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    tmp = (PyObject *)self->loop;
+    tmp = (PyObject *)((Handle *)self)->loop;
     Py_INCREF(loop);
-    self->loop = loop;
+    ((Handle *)self)->loop = loop;
     Py_XDECREF(tmp);
 
     uv_prepare = PyMem_Malloc(sizeof(uv_prepare_t));
@@ -197,14 +113,14 @@ Prepare_tp_init(Prepare *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    r = uv_prepare_init(UV_LOOP(self), uv_prepare);
+    r = uv_prepare_init(UV_HANDLE_LOOP(self), uv_prepare);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_PrepareError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_PrepareError);
         Py_DECREF(loop);
         return -1;
     }
     uv_prepare->data = (void *)self;
-    self->uv_handle = uv_prepare;
+    UV_HANDLE(self) = (uv_handle_t *)uv_prepare;
 
     return 0;
 }
@@ -213,11 +129,10 @@ Prepare_tp_init(Prepare *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Prepare_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Prepare *self = (Prepare *)PyType_GenericNew(type, args, kwargs);
+    Prepare *self = (Prepare *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
         return NULL;
     }
-    self->uv_handle = NULL;
     return (PyObject *)self;
 }
 
@@ -225,10 +140,8 @@ Prepare_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 Prepare_tp_traverse(Prepare *self, visitproc visit, void *arg)
 {
-    Py_VISIT(self->data);
     Py_VISIT(self->callback);
-    Py_VISIT(self->on_close_cb);
-    Py_VISIT(self->loop);
+    HandleType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
 
@@ -236,54 +149,26 @@ Prepare_tp_traverse(Prepare *self, visitproc visit, void *arg)
 static int
 Prepare_tp_clear(Prepare *self)
 {
-    Py_CLEAR(self->data);
     Py_CLEAR(self->callback);
-    Py_CLEAR(self->on_close_cb);
-    Py_CLEAR(self->loop);
+    HandleType.tp_clear((PyObject *)self);
     return 0;
 }
 
 
-static void
-Prepare_tp_dealloc(Prepare *self)
-{
-    if (self->uv_handle) {
-        uv_close((uv_handle_t *)self->uv_handle, on_prepare_dealloc_close);
-        self->uv_handle = NULL;
-    }
-    Prepare_tp_clear(self);
-    Py_TYPE(self)->tp_free((PyObject *)self);
-}
-
-
 static PyMethodDef
 Prepare_tp_methods[] = {
     { "start", (PyCFunction)Prepare_func_start, METH_VARARGS, "Start the Prepare." },
     { "stop", (PyCFunction)Prepare_func_stop, METH_NOARGS, "Stop the Prepare." },
-    { "close", (PyCFunction)Prepare_func_close, METH_VARARGS, "Close the Prepare." },
     { NULL }
 };
 
 
-static PyMemberDef Prepare_tp_members[] = {
-    {"loop", T_OBJECT_EX, offsetof(Prepare, loop), READONLY, "Loop where this Prepare is running on."},
-    {"data", T_OBJECT, offsetof(Prepare, data), 0, "Arbitrary data."},
-    {NULL}
-};
-
-
-static PyGetSetDef Prepare_tp_getsets[] = {
-    {"active", (getter)Prepare_active_get, 0, "Indicates if handle is active", NULL},
-    {NULL}
-};
-
-
 static PyTypeObject PrepareType = {
     PyVarObject_HEAD_INIT(NULL, 0)
     "pyuv.Prepare",                                                 /*tp_name*/
     sizeof(Prepare),                                                /*tp_basicsize*/
     0,                                                              /*tp_itemsize*/
-    (destructor)Prepare_tp_dealloc,                                 /*tp_dealloc*/
+    0,                                                              /*tp_dealloc*/
     0,                                                              /*tp_print*/
     0,                                                              /*tp_getattr*/
     0,                                                              /*tp_setattr*/
@@ -307,8 +192,8 @@ static PyTypeObject PrepareType = {
     0,                                                              /*tp_iter*/
     0,                                                              /*tp_iternext*/
     Prepare_tp_methods,                                             /*tp_methods*/
-    Prepare_tp_members,                                             /*tp_members*/
-    Prepare_tp_getsets,                                             /*tp_getsets*/
+    0,                                                              /*tp_members*/
+    0,                                                              /*tp_getsets*/
     0,                                                              /*tp_base*/
     0,                                                              /*tp_dict*/
     0,                                                              /*tp_descr_get*/
