@@ -8,6 +8,176 @@
  * http://stackoverflow.com/questions/2055918/forcing-a-program-to-flush-its-standard-output-when-redirected
  */
 
+/* Container for standard IO */
+static int
+StdIO_tp_init(StdIO *self, PyObject *args, PyObject *kwargs)
+{
+    int flags = UV_IGNORE;
+    int fd = -1;
+    PyObject *stream, *tmp;
+
+    static char *kwlist[] = {"stream", "fd", "flags", NULL};
+
+    stream = tmp = NULL;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|Oii:__init__", kwlist, &stream, &fd, &flags)) {
+        return -1;
+    }
+
+    if (stream != NULL && fd != -1) {
+        PyErr_SetString(PyExc_ValueError, "either stream or fd must be specified, but not both");
+        return -1;
+    }
+
+    if (stream != NULL) {
+        if (!PyObject_IsSubclass((PyObject *)stream->ob_type, (PyObject *)&StreamType)) {
+            PyErr_SetString(PyExc_TypeError, "Only stream objects are supported");
+            return -1;
+        }
+        if (!((flags & ~(UV_CREATE_PIPE | UV_READABLE_PIPE | UV_WRITABLE_PIPE | UV_INHERIT_STREAM))==0)) {
+            PyErr_SetString(PyExc_ValueError, "invalid flags specified for stream");
+            return -1;
+        }
+    }
+
+    if (fd != -1 && !((flags & ~(UV_INHERIT_FD))==0)) {
+        PyErr_SetString(PyExc_ValueError, "invalid flags specified for fd");
+        return -1;
+    }
+
+    if (stream == NULL && fd == -1 && !((flags & ~(UV_IGNORE))==0)) {
+        PyErr_SetString(PyExc_ValueError, "invalid flags specified for ignore");
+        return -1;
+    }
+
+    tmp = self->stream;
+    Py_XINCREF(stream);
+    self->stream = stream;
+    Py_XDECREF(tmp);
+
+    self->fd = fd;
+    self->flags = flags;
+
+    return 0;
+}
+
+
+static PyObject *
+StdIO_stream_get(StdIO *self, void *closure)
+{
+    UNUSED_ARG(closure);
+    if (!self->stream) {
+        Py_INCREF(Py_None);
+        Py_RETURN_NONE;
+    } else {
+        return self->stream;
+    }
+}
+
+
+static PyObject *
+StdIO_fd_get(StdIO *self, void *closure)
+{
+    UNUSED_ARG(closure);
+    return PyInt_FromLong((long)self->fd);
+}
+
+
+static PyObject *
+StdIO_flags_get(StdIO *self, void *closure)
+{
+    UNUSED_ARG(closure);
+    return PyInt_FromLong((long)self->fd);
+}
+
+
+static PyObject *
+StdIO_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
+{
+    StdIO *self = (StdIO *)PyType_GenericNew(type, args, kwargs);
+    if (!self) {
+        return NULL;
+    }
+    return (PyObject *)self;
+}
+
+
+static int
+StdIO_tp_traverse(StdIO *self, visitproc visit, void *arg)
+{
+    Py_VISIT(self->stream);
+    return 0;
+}
+
+
+static int
+StdIO_tp_clear(StdIO *self)
+{
+    Py_CLEAR(self->stream);
+    return 0;
+}
+
+
+static void
+StdIO_tp_dealloc(StdIO *self)
+{
+    StdIO_tp_clear(self);
+    Py_TYPE(self)->tp_free((PyObject *)self);
+}
+
+
+static PyGetSetDef StdIO_tp_getsets[] = {
+    {"stream", (getter)StdIO_stream_get, NULL, "Stream object.", NULL},
+    {"fd", (getter)StdIO_fd_get, NULL, "File descriptor.", NULL},
+    {"flags", (getter)StdIO_flags_get, NULL, "Flags.", NULL},
+    {NULL}
+};
+
+
+static PyTypeObject StdIOType = {
+    PyVarObject_HEAD_INIT(NULL, 0)
+    "pyuv.StdIO",                                                   /*tp_name*/
+    sizeof(StdIO),                                                  /*tp_basicsize*/
+    0,                                                              /*tp_itemsize*/
+    (destructor)StdIO_tp_dealloc,                                   /*tp_dealloc*/
+    0,                                                              /*tp_print*/
+    0,                                                              /*tp_getattr*/
+    0,                                                              /*tp_setattr*/
+    0,                                                              /*tp_compare*/
+    0,                                                              /*tp_repr*/
+    0,                                                              /*tp_as_number*/
+    0,                                                              /*tp_as_sequence*/
+    0,                                                              /*tp_as_mapping*/
+    0,                                                              /*tp_hash */
+    0,                                                              /*tp_call*/
+    0,                                                              /*tp_str*/
+    0,                                                              /*tp_getattro*/
+    0,                                                              /*tp_setattro*/
+    0,                                                              /*tp_as_buffer*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,                        /*tp_flags*/
+    0,                                                              /*tp_doc*/
+    (traverseproc)StdIO_tp_traverse,                                /*tp_traverse*/
+    (inquiry)StdIO_tp_clear,                                        /*tp_clear*/
+    0,                                                              /*tp_richcompare*/
+    0,                                                              /*tp_weaklistoffset*/
+    0,                                                              /*tp_iter*/
+    0,                                                              /*tp_iternext*/
+    0,                                                              /*tp_methods*/
+    0,                                                              /*tp_members*/
+    StdIO_tp_getsets,                                               /*tp_getsets*/
+    0,                                                              /*tp_base*/
+    0,                                                              /*tp_dict*/
+    0,                                                              /*tp_descr_get*/
+    0,                                                              /*tp_descr_set*/
+    0,                                                              /*tp_dictoffset*/
+    (initproc)StdIO_tp_init,                                        /*tp_init*/
+    0,                                                              /*tp_alloc*/
+    StdIO_tp_new,                                                   /*tp_new*/
+};
+
+
+/* Process handle */
+
 static PyObject* PyExc_ProcessError;
 
 
@@ -44,72 +214,33 @@ on_process_exit(uv_process_t *process, int exit_status, int term_signal)
 }
 
 
-static void
-on_process_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    Process *self;
-    PyObject *result;
-
-    ASSERT(handle);
-    self = (Process *)handle->data;
-    ASSERT(self);
-
-    if (self->on_close_cb != Py_None) {
-        result = PyObject_CallFunctionObjArgs(self->on_close_cb, self, NULL);
-        if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_close_cb);
-        }
-        Py_XDECREF(result);
-    }
-
-    handle->data = NULL;
-    PyMem_Free(handle);
-
-    /* Refcount was increased in func_close */
-    Py_DECREF(self);
-
-    PyGILState_Release(gstate);
-}
-
-
-static void
-on_process_dealloc_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    ASSERT(handle);
-    handle->data = NULL;
-    PyMem_Free(handle);
-    PyGILState_Release(gstate);
-}
-
-
 static PyObject *
 Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
 {
-    int r, flags;
+    int r, flags, stdio_count;
     unsigned int uid, gid;
     char *cwd, *cwd2, *file, *file2, *arg_str, *tmp_str, *key_str, *value_str;
     char **ptr, **process_args, **process_env;
     Py_ssize_t i, n, pos;
-    PyObject *key, *value, *item, *tmp, *callback, *arguments, *env, *stdin_pipe, *stdout_pipe, *stderr_pipe;
+    PyObject *key, *value, *item, *tmp, *callback, *arguments, *env, *stdio;
     uv_process_t *uv_process;
     uv_process_options_t options;
+    uv_stdio_container_t *stdio_container;
 
-    static char *kwlist[] = {"file", "exit_callback", "args", "env", "cwd", "uid", "gid", "flags", "stdin", "stdout", "stderr", NULL};
+    static char *kwlist[] = {"file", "exit_callback", "args", "env", "cwd", "uid", "gid", "flags", "stdio", NULL};
 
     cwd = NULL;
     ptr = process_args = process_env = NULL;
-    tmp = arguments = env = NULL;
-    stdin_pipe = stdout_pipe = stderr_pipe = Py_None;
-    flags = uid = gid = 0;
+    tmp = arguments = env = stdio = NULL;
+    stdio_container = NULL;
+    flags = uid = gid = stdio_count = 0;
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_ProcessError, "Process already spawned");
         return NULL;
     }
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|OOO!sIIiO!O!O!:__init__", kwlist, &file, &callback, &arguments, &PyDict_Type, &env, &cwd, &uid, &gid, &flags, &PipeType, &stdin_pipe, &PipeType, &stdout_pipe, &PipeType, &stderr_pipe)) {
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|OOO!sIIiO:__init__", kwlist, &file, &callback, &arguments, &PyDict_Type, &env, &cwd, &uid, &gid, &flags, &stdio)) {
         return NULL;
     }
 
@@ -119,7 +250,12 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     }
 
     if (arguments && !PySequence_Check(arguments)) {
-        PyErr_SetString(PyExc_TypeError, "only iterable objects are supported");
+        PyErr_SetString(PyExc_TypeError, "only iterable objects are supported for 'args'");
+        return NULL;
+    }
+
+    if (stdio && !PySequence_Check(stdio)) {
+        PyErr_SetString(PyExc_TypeError, "only iterable objects are supported for 'stdio'");
         return NULL;
     }
 
@@ -128,19 +264,9 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     self->on_exit_cb = callback;
     Py_XDECREF(tmp);
 
-    tmp = (PyObject *)self->stdin_pipe;
-    Py_INCREF(stdin_pipe);
-    self->stdin_pipe = stdin_pipe;
-    Py_XDECREF(tmp);
-
-    tmp = (PyObject *)self->stdout_pipe;
-    Py_INCREF(stdout_pipe);
-    self->stdout_pipe = stdout_pipe;
-    Py_XDECREF(tmp);
-
-    tmp = (PyObject *)self->stderr_pipe;
-    Py_INCREF(stderr_pipe);
-    self->stderr_pipe = stderr_pipe;
+    tmp = self->stdio;
+    Py_XINCREF(stdio);
+    self->stdio = stdio;
     Py_XDECREF(tmp);
 
     memset(&options, 0, sizeof(uv_process_options_t));
@@ -229,14 +355,32 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
     }
     options.env = process_env;
 
-    if (stdin_pipe != Py_None)
-        options.stdin_stream = (uv_pipe_t *)((IOStream *)stdin_pipe)->uv_handle;
-
-    if (stdout_pipe != Py_None)
-        options.stdout_stream = (uv_pipe_t *)((IOStream *)stdout_pipe)->uv_handle;
-
-    if (stderr_pipe != Py_None)
-        options.stderr_stream = (uv_pipe_t *)((IOStream *)stderr_pipe)->uv_handle;
+    if (stdio) {
+        n = PySequence_Length(stdio);
+        stdio_container = (uv_stdio_container_t *)PyMem_Malloc(sizeof(uv_stdio_container_t) * n);
+        if (!stdio_container) {
+            PyErr_NoMemory();
+            goto error;
+        }
+        item = NULL;
+        for (i = 0;i < n; i++) {
+            item = PySequence_GetItem(stdio, i);
+            if (!item || !PyObject_TypeCheck(item, &StdIOType)) {
+                Py_XDECREF(item);
+                continue;
+            }
+            stdio_count++;
+            stdio_container[i].flags = ((StdIO *)item)->flags;
+            if (((StdIO *)item)->flags & (UV_CREATE_PIPE | UV_INHERIT_STREAM)) {
+                stdio_container[i].data.stream = (uv_stream_t *)(UV_HANDLE(((StdIO *)item)->stream));
+            } else if (((StdIO *)item)->flags & UV_INHERIT_FD) {
+                stdio_container[i].data.fd = ((StdIO *)item)->fd;
+            }
+            Py_DECREF(item);
+        }
+    }
+    options.stdio = stdio_container;
+    options.stdio_count = stdio_count;
 
     uv_process = PyMem_Malloc(sizeof(uv_process_t));
     if (!uv_process) {
@@ -244,9 +388,9 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
         goto error;
     }
     uv_process->data = (void *)self;
-    self->uv_handle = uv_process;
+    UV_HANDLE(self) = (uv_handle_t *)uv_process;
 
-    r = uv_spawn(UV_LOOP(self), self->uv_handle, options);
+    r = uv_spawn(UV_HANDLE_LOOP(self), uv_process, options);
 
     if (options.args) {
         for (ptr = options.args; *ptr != NULL; ptr++) {
@@ -263,11 +407,14 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
         }
         PyMem_Free(options.env);
     }
+    if (options.stdio) {
+        PyMem_Free(options.stdio);
+    }
 
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_ProcessError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_ProcessError);
         PyMem_Free(uv_process);
-        self->uv_handle = NULL;
+        UV_HANDLE(self) = NULL;
         goto error;
     }
 
@@ -275,9 +422,7 @@ Process_func_spawn(Process *self, PyObject *args, PyObject *kwargs)
 
 error:
     Py_DECREF(self->on_exit_cb);
-    Py_DECREF(self->stdin_pipe);
-    Py_DECREF(self->stdout_pipe);
-    Py_DECREF(self->stderr_pipe);
+    Py_XDECREF(self->stdio);
     return NULL;
 
 }
@@ -288,7 +433,7 @@ Process_func_kill(Process *self, PyObject *args)
 {
     int signum, r;
 
-    if (!self->uv_handle) {
+    if (UV_HANDLE_CLOSED(self)) {
         PyErr_SetString(PyExc_ProcessError, "Process wasn't spawned yet");
         return NULL;
     }
@@ -297,9 +442,9 @@ Process_func_kill(Process *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_process_kill(self->uv_handle, signum);
+    r = uv_process_kill((uv_process_t *)UV_HANDLE(self), signum);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_ProcessError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_ProcessError);
         return NULL;
     }
 
@@ -307,47 +452,15 @@ Process_func_kill(Process *self, PyObject *args)
 }
 
 
-static PyObject *
-Process_func_close(Process *self, PyObject *args)
-{
-    PyObject *callback = Py_None;
-
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_ProcessError, "Process wasn't spawned yet");
-        return NULL;
-    }
-
-    if (!PyArg_ParseTuple(args, "|O:close", &callback)) {
-        return NULL;
-    }
-
-    if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    Py_INCREF(callback);
-    self->on_close_cb = callback;
-
-    /* Increase refcount so that object is not removed before the callback is called */
-    Py_INCREF(self);
-
-    uv_close((uv_handle_t *)self->uv_handle, on_process_close);
-    self->uv_handle = NULL;
-
-    Py_RETURN_NONE;
-}
-
-
 static PyObject *
 Process_pid_get(Process *self, void *closure)
 {
     UNUSED_ARG(closure);
 
-    if (!self->uv_handle) {
+    if (UV_HANDLE_CLOSED(self)) {
         Py_RETURN_NONE;
     }
-    return PyInt_FromLong((long)self->uv_handle->pid);
+    return PyInt_FromLong((long)((uv_process_t *)UV_HANDLE(self))->pid);
 }
 
 
@@ -359,7 +472,7 @@ Process_tp_init(Process *self, PyObject *args, PyObject *kwargs)
 
     UNUSED_ARG(kwargs);
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_ProcessError, "Object already initialized");
         return -1;
     }
@@ -368,9 +481,9 @@ Process_tp_init(Process *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    tmp = (PyObject *)self->loop;
+    tmp = (PyObject *)((Handle *)self)->loop;
     Py_INCREF(loop);
-    self->loop = loop;
+    ((Handle *)self)->loop = loop;
     Py_XDECREF(tmp);
 
     return 0;
@@ -380,11 +493,10 @@ Process_tp_init(Process *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Process_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Process *self = (Process *)PyType_GenericNew(type, args, kwargs);
+    Process *self = (Process *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
         return NULL;
     }
-    self->uv_handle = NULL;
     return (PyObject *)self;
 }
 
@@ -392,13 +504,9 @@ Process_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 Process_tp_traverse(Process *self, visitproc visit, void *arg)
 {
-    Py_VISIT(self->data);
-    Py_VISIT(self->on_close_cb);
     Py_VISIT(self->on_exit_cb);
-    Py_VISIT(self->stdin_pipe);
-    Py_VISIT(self->stdout_pipe);
-    Py_VISIT(self->stderr_pipe);
-    Py_VISIT(self->loop);
+    Py_VISIT(self->stdio);
+    HandleType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
 
@@ -406,45 +514,21 @@ Process_tp_traverse(Process *self, visitproc visit, void *arg)
 static int
 Process_tp_clear(Process *self)
 {
-    Py_CLEAR(self->data);
-    Py_CLEAR(self->on_close_cb);
     Py_CLEAR(self->on_exit_cb);
-    Py_CLEAR(self->stdin_pipe);
-    Py_CLEAR(self->stdout_pipe);
-    Py_CLEAR(self->stderr_pipe);
-    Py_CLEAR(self->loop);
+    Py_CLEAR(self->stdio);
+    HandleType.tp_clear((PyObject *)self);
     return 0;
 }
 
 
-static void
-Process_tp_dealloc(Process *self)
-{
-    if (self->uv_handle) {
-        uv_close((uv_handle_t *)self->uv_handle, on_process_dealloc_close);
-        self->uv_handle = NULL;
-    }
-    Process_tp_clear(self);
-    Py_TYPE(self)->tp_free((PyObject *)self);
-}
-
-
 static PyMethodDef
 Process_tp_methods[] = {
     { "spawn", (PyCFunction)Process_func_spawn, METH_VARARGS|METH_KEYWORDS, "Spawn the child process." },
     { "kill", (PyCFunction)Process_func_kill, METH_VARARGS, "Kill this process with the specified signal number." },
-    { "close", (PyCFunction)Process_func_close, METH_VARARGS, "Close the Process." },
     { NULL }
 };
 
 
-static PyMemberDef Process_tp_members[] = {
-    {"loop", T_OBJECT_EX, offsetof(Process, loop), READONLY, "Loop where this Process is running on."},
-    {"data", T_OBJECT, offsetof(Process, data), 0, "Arbitrary data."},
-    {NULL}
-};
-
-
 static PyGetSetDef Process_tp_getsets[] = {
     {"pid", (getter)Process_pid_get, NULL, "Process ID", NULL},
     {NULL}
@@ -456,7 +540,7 @@ static PyTypeObject ProcessType = {
     "pyuv.Process",                                                 /*tp_name*/
     sizeof(Process),                                                /*tp_basicsize*/
     0,                                                              /*tp_itemsize*/
-    (destructor)Process_tp_dealloc,                                 /*tp_dealloc*/
+    0,                                                              /*tp_dealloc*/
     0,                                                              /*tp_print*/
     0,                                                              /*tp_getattr*/
     0,                                                              /*tp_setattr*/
@@ -480,7 +564,7 @@ static PyTypeObject ProcessType = {
     0,                                                              /*tp_iter*/
     0,                                                              /*tp_iternext*/
     Process_tp_methods,                                             /*tp_methods*/
-    Process_tp_members,                                             /*tp_members*/
+    0,                                                              /*tp_members*/
     Process_tp_getsets,                                             /*tp_getsets*/
     0,                                                              /*tp_base*/
     0,                                                              /*tp_dict*/
