@@ -2,47 +2,6 @@
 static PyObject* PyExc_TimerError;
 
 
-static void
-on_timer_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    Timer *self;
-    PyObject *result;
-
-    ASSERT(handle);
-
-    self = (Timer *)handle->data;
-    ASSERT(self);
-
-    if (self->on_close_cb != Py_None) {
-        result = PyObject_CallFunctionObjArgs(self->on_close_cb, self, NULL);
-        if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_close_cb);
-        }
-        Py_XDECREF(result);
-    }
-
-    handle->data = NULL;
-    PyMem_Free(handle);
-
-    /* Refcount was increased in func_close */
-    Py_DECREF(self);
-
-    PyGILState_Release(gstate);
-}
-
-
-static void
-on_timer_dealloc_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    ASSERT(handle);
-    handle->data = NULL;
-    PyMem_Free(handle);
-    PyGILState_Release(gstate);
-}
-
-
 static void
 on_timer_callback(uv_timer_t *timer, int status)
 {
@@ -80,10 +39,7 @@ Timer_func_start(Timer *self, PyObject *args, PyObject *kwargs)
 
     tmp = NULL;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_TimerError, "Timer is closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs, "Odd:__init__", kwlist, &callback, &timeout, &repeat)) {
         return NULL;
@@ -104,9 +60,9 @@ Timer_func_start(Timer *self, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
-    r = uv_timer_start(self->uv_handle, on_timer_callback, (int64_t)(timeout * 1000), (int64_t)(repeat * 1000));
+    r = uv_timer_start((uv_timer_t *)UV_HANDLE(self), on_timer_callback, (int64_t)(timeout * 1000), (int64_t)(repeat * 1000));
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_TimerError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TimerError);
         return NULL;
     }
 
@@ -124,14 +80,11 @@ Timer_func_stop(Timer *self)
 {
     int r;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_TimerError, "Timer is closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_timer_stop(self->uv_handle);
+    r = uv_timer_stop((uv_timer_t *)UV_HANDLE(self));
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_TimerError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TimerError);
         return NULL;
     }
 
@@ -144,14 +97,11 @@ Timer_func_again(Timer *self)
 {
     int r;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_TimerError, "Timer is closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_timer_again(self->uv_handle);
+    r = uv_timer_again((uv_timer_t *)UV_HANDLE(self));
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_TimerError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TimerError);
         return NULL;
     }
 
@@ -159,59 +109,12 @@ Timer_func_again(Timer *self)
 }
 
 
-static PyObject *
-Timer_func_close(Timer *self, PyObject *args)
-{
-    PyObject *callback = Py_None;
-
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_TimerError, "Timer is already closed");
-        return NULL;
-    }
-
-    if (!PyArg_ParseTuple(args, "|O:close", &callback)) {
-        return NULL;
-    }
-
-    if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    Py_INCREF(callback);
-    self->on_close_cb = callback;
-
-    /* Increase refcount so that object is not removed before the callback is called */
-    Py_INCREF(self);
-
-    uv_close((uv_handle_t *)self->uv_handle, on_timer_close);
-    self->uv_handle = NULL;
-
-    Py_RETURN_NONE;
-}
-
-
-static PyObject *
-Timer_active_get(Timer *self, void *closure)
-{
-    UNUSED_ARG(closure);
-    if (!self->uv_handle) {
-        Py_RETURN_FALSE;
-    } else {
-        return PyBool_FromLong((long)uv_is_active((uv_handle_t *)self->uv_handle));
-    }
-}
-
-
 static PyObject *
 Timer_repeat_get(Timer *self, void *closure)
 {
     UNUSED_ARG(closure);
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_TimerError, "Timer is closed");
-        return NULL;
-    }
-    return PyFloat_FromDouble(uv_timer_get_repeat(self->uv_handle)/1000.0);
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+    return PyFloat_FromDouble(uv_timer_get_repeat((uv_timer_t *)UV_HANDLE(self))/1000.0);
 }
 
 
@@ -222,10 +125,7 @@ Timer_repeat_set(Timer *self, PyObject *value, void *closure)
 
     UNUSED_ARG(closure);
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_TimerError, "Timer is closed");
-        return -1;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, -1);
 
     if (!value) {
         PyErr_SetString(PyExc_TypeError, "cannot delete attribute");
@@ -242,7 +142,7 @@ Timer_repeat_set(Timer *self, PyObject *value, void *closure)
         return -1;
     }
 
-    uv_timer_set_repeat(self->uv_handle, (int64_t)(repeat * 1000));
+    uv_timer_set_repeat((uv_timer_t *)UV_HANDLE(self), (int64_t)(repeat * 1000));
 
     return 0;
 }
@@ -258,7 +158,7 @@ Timer_tp_init(Timer *self, PyObject *args, PyObject *kwargs)
 
     UNUSED_ARG(kwargs);
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_TimerError, "Object already initialized");
         return -1;
     }
@@ -267,9 +167,9 @@ Timer_tp_init(Timer *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    tmp = (PyObject *)self->loop;
+    tmp = (PyObject *)((Handle *)self)->loop;
     Py_INCREF(loop);
-    self->loop = loop;
+    ((Handle *)self)->loop = loop;
     Py_XDECREF(tmp);
 
     uv_timer = PyMem_Malloc(sizeof(uv_timer_t));
@@ -278,15 +178,15 @@ Timer_tp_init(Timer *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    r = uv_timer_init(UV_LOOP(self), uv_timer);
+    r = uv_timer_init(UV_HANDLE_LOOP(self), uv_timer);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_TimerError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_TimerError);
         PyMem_Free(uv_timer);
         Py_DECREF(loop);
         return -1;
     }
     uv_timer->data = (void *)self;
-    self->uv_handle = uv_timer;
+    UV_HANDLE(self) = (uv_handle_t *)uv_timer;
 
     return 0;
 }
@@ -295,11 +195,10 @@ Timer_tp_init(Timer *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 Timer_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    Timer *self = (Timer *)PyType_GenericNew(type, args, kwargs);
+    Timer *self = (Timer *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
         return NULL;
     }
-    self->uv_handle = NULL;
     return (PyObject *)self;
 }
 
@@ -307,10 +206,8 @@ Timer_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 Timer_tp_traverse(Timer *self, visitproc visit, void *arg)
 {
-    Py_VISIT(self->data);
     Py_VISIT(self->callback);
-    Py_VISIT(self->on_close_cb);
-    Py_VISIT(self->loop);
+    HandleType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
 
@@ -318,44 +215,22 @@ Timer_tp_traverse(Timer *self, visitproc visit, void *arg)
 static int
 Timer_tp_clear(Timer *self)
 {
-    Py_CLEAR(self->data);
     Py_CLEAR(self->callback);
-    Py_CLEAR(self->on_close_cb);
-    Py_CLEAR(self->loop);
+    HandleType.tp_clear((PyObject *)self);
     return 0;
 }
 
 
-static void
-Timer_tp_dealloc(Timer *self)
-{
-    if (self->uv_handle) {
-        uv_close((uv_handle_t *)self->uv_handle, on_timer_dealloc_close);
-    }
-    Timer_tp_clear(self);
-    Py_TYPE(self)->tp_free((PyObject *)self);
-}
-
-
 static PyMethodDef
 Timer_tp_methods[] = {
     { "start", (PyCFunction)Timer_func_start, METH_VARARGS|METH_KEYWORDS, "Start the Timer." },
     { "stop", (PyCFunction)Timer_func_stop, METH_NOARGS, "Stop the Timer." },
     { "again", (PyCFunction)Timer_func_again, METH_NOARGS, "Stop the timer, and if it is repeating restart it using the repeat value as the timeout." },
-    { "close", (PyCFunction)Timer_func_close, METH_VARARGS, "Close the Timer." },
     { NULL }
 };
 
 
-static PyMemberDef Timer_tp_members[] = {
-    {"loop", T_OBJECT_EX, offsetof(Timer, loop), READONLY, "Loop where this Timer is running on."},
-    {"data", T_OBJECT, offsetof(Timer, data), 0, "Arbitrary data."},
-    {NULL}
-};
-
-
 static PyGetSetDef Timer_tp_getsets[] = {
-    {"active", (getter)Timer_active_get, 0, "Indicates if handle is active.", NULL},
     {"repeat", (getter)Timer_repeat_get, (setter)Timer_repeat_set, "Timer repeat value.", NULL},
     {NULL}
 };
@@ -366,7 +241,7 @@ static PyTypeObject TimerType = {
     "pyuv.Timer",                                                   /*tp_name*/
     sizeof(Timer),                                                  /*tp_basicsize*/
     0,                                                              /*tp_itemsize*/
-    (destructor)Timer_tp_dealloc,                                   /*tp_dealloc*/
+    0,                                                              /*tp_dealloc*/
     0,                                                              /*tp_print*/
     0,                                                              /*tp_getattr*/
     0,                                                              /*tp_setattr*/
@@ -390,7 +265,7 @@ static PyTypeObject TimerType = {
     0,                                                              /*tp_iter*/
     0,                                                              /*tp_iternext*/
     Timer_tp_methods,                                               /*tp_methods*/
-    Timer_tp_members,                                               /*tp_members*/
+    0,                                                              /*tp_members*/
     Timer_tp_getsets,                                               /*tp_getsets*/
     0,                                                              /*tp_base*/
     0,                                                              /*tp_dict*/
