@@ -1,68 +1,24 @@
 
-#define UDP_MAX_BUF_SIZE 65536
-
 static PyObject* PyExc_UDPError;
 
 
-typedef struct {
-    PyObject *obj;
-    PyObject *callback;
-    void *data;
-} udp_req_data_t;
-
 typedef struct {
     uv_buf_t *bufs;
     int buf_count;
+    Py_buffer view;
 } udp_send_data_t;
 
-
-static void
-on_udp_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    UDP *self;
-    PyObject *result;
-
-    ASSERT(handle);
-    self = (UDP *)handle->data;
-    ASSERT(self);
-
-    if (self->on_close_cb != Py_None) {
-        result = PyObject_CallFunctionObjArgs(self->on_close_cb, self, NULL);
-        if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_close_cb);
-        }
-        Py_XDECREF(result);
-    }
-
-    handle->data = NULL;
-    PyMem_Free(handle);
-
-    /* Refcount was increased in func_close */
-    Py_DECREF(self);
-
-    PyGILState_Release(gstate);
-}
-
-
-static void
-on_udp_dealloc_close(uv_handle_t *handle)
-{
-    PyGILState_STATE gstate = PyGILState_Ensure();
-    ASSERT(handle);
-    handle->data = NULL;
-    PyMem_Free(handle);
-    PyGILState_Release(gstate);
-}
+typedef struct {
+    PyObject *callback;
+    udp_send_data_t *data;
+} udp_req_data_t;
 
 
 static uv_buf_t
 on_udp_alloc(uv_udp_t* handle, size_t suggested_size)
 {
     PyGILState_STATE gstate = PyGILState_Ensure();
-    uv_buf_t buf;
-    ASSERT(suggested_size <= UDP_MAX_BUF_SIZE);
-    buf = uv_buf_init(PyMem_Malloc(suggested_size), suggested_size);
+    uv_buf_t buf = uv_buf_init(PyMem_Malloc(suggested_size), suggested_size);
     UNUSED_ARG(handle);
     PyGILState_Release(gstate);
     return buf;
@@ -89,6 +45,10 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
     /* Object could go out of scope in the callback, increase refcount to avoid it */
     Py_INCREF(self);
 
+    if (nread == 0) {
+        goto done;
+    }
+
     if (nread > 0) {
         ASSERT(addr);
         if (addr->sa_family == AF_INET) {
@@ -108,21 +68,20 @@ on_udp_read(uv_udp_t* handle, int nread, uv_buf_t buf, struct sockaddr* addr, un
         Py_INCREF(Py_None);
         data = Py_None;
         Py_INCREF(Py_None);
-        err = uv_last_error(UV_LOOP(self));
+        err = uv_last_error(UV_HANDLE_LOOP(self));
         py_errorno = PyInt_FromLong((long)err.code);
     }
 
-    if (nread != 0) {
-        result = PyObject_CallFunctionObjArgs(self->on_read_cb, self, address_tuple, data, py_errorno, NULL);
-        if (result == NULL) {
-            PyErr_WriteUnraisable(self->on_read_cb);
-        }
-        Py_XDECREF(result);
-        Py_DECREF(address_tuple);
-        Py_DECREF(data);
-        Py_DECREF(py_errorno);
+    result = PyObject_CallFunctionObjArgs(self->on_read_cb, self, address_tuple, data, py_errorno, NULL);
+    if (result == NULL) {
+        PyErr_WriteUnraisable(self->on_read_cb);
     }
+    Py_XDECREF(result);
+    Py_DECREF(address_tuple);
+    Py_DECREF(data);
+    Py_DECREF(py_errorno);
 
+done:
     PyMem_Free(buf.base);
 
     Py_DECREF(self);
@@ -143,9 +102,9 @@ on_udp_send(uv_udp_send_t* req, int status)
     ASSERT(req);
 
     req_data = (udp_req_data_t *)req->data;
-    send_data = (udp_send_data_t *)req_data->data;
+    send_data = req_data->data;
 
-    self = (UDP *)req_data->obj;
+    self = (UDP *)req->handle->data;
     callback = req_data->callback;
 
     ASSERT(self);
@@ -154,7 +113,7 @@ on_udp_send(uv_udp_send_t* req, int status)
 
     if (callback != Py_None) {
         if (status < 0) {
-            uv_err_t err = uv_last_error(UV_LOOP(self));
+            uv_err_t err = uv_last_error(UV_HANDLE_LOOP(self));
             py_errorno = PyInt_FromLong((long)err.code);
         } else {
             py_errorno = Py_None;
@@ -168,8 +127,12 @@ on_udp_send(uv_udp_send_t* req, int status)
         Py_DECREF(py_errorno);
     }
 
-    for (i = 0; i < send_data->buf_count; i++) {
-        PyMem_Free(send_data->bufs[i].base);
+    if (send_data->buf_count == 1) {
+        PyBuffer_Release(&send_data->view);
+    } else {
+        for (i = 0; i < send_data->buf_count; i++) {
+            PyMem_Free(send_data->bufs[i].base);
+        }
     }
     PyMem_Free(send_data);
     Py_DECREF(callback);
@@ -189,10 +152,7 @@ UDP_func_bind(UDP *self, PyObject *args)
     struct in_addr addr4;
     struct in6_addr addr6;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "(si):bind", &bind_ip, &bind_port)) {
         return NULL;
@@ -213,13 +173,13 @@ UDP_func_bind(UDP *self, PyObject *args)
     }
 
     if (address_type == AF_INET) {
-        r = uv_udp_bind(self->uv_handle, uv_ip4_addr(bind_ip, bind_port), 0);
+        r = uv_udp_bind((uv_udp_t *)UV_HANDLE(self), uv_ip4_addr(bind_ip, bind_port), 0);
     } else {
-        r = uv_udp_bind6(self->uv_handle, uv_ip6_addr(bind_ip, bind_port), UV_UDP_IPV6ONLY);
+        r = uv_udp_bind6((uv_udp_t *)UV_HANDLE(self), uv_ip6_addr(bind_ip, bind_port), UV_UDP_IPV6ONLY);
     }
 
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -235,10 +195,7 @@ UDP_func_start_recv(UDP *self, PyObject *args)
 
     tmp = NULL;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O:start_recv", &callback)) {
         return NULL;
@@ -249,9 +206,9 @@ UDP_func_start_recv(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_recv_start(self->uv_handle, (uv_alloc_cb)on_udp_alloc, (uv_udp_recv_cb)on_udp_read);
+    r = uv_udp_recv_start((uv_udp_t *)UV_HANDLE(self), (uv_alloc_cb)on_udp_alloc, (uv_udp_recv_cb)on_udp_read);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -269,16 +226,17 @@ UDP_func_stop_recv(UDP *self)
 {
     int r;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+
+    r = uv_udp_recv_stop((uv_udp_t *)UV_HANDLE(self));
+    if (r != 0) {
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
-    r = uv_udp_recv_stop(self->uv_handle);
-    if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
-        return NULL;
-    }
+    Py_XDECREF(self->on_read_cb);
+    self->on_read_cb = NULL;
+
     Py_RETURN_NONE;
 }
 
@@ -286,32 +244,138 @@ UDP_func_stop_recv(UDP *self)
 static PyObject *
 UDP_func_send(UDP *self, PyObject *args)
 {
-    int i, n, r, buf_count, dest_port, address_type;
-    char *data_str, *tmp, *dest_ip;
+    int r, buf_count, dest_port, address_type;
+    char *dest_ip;
     struct in_addr addr4;
     struct in6_addr addr6;
-    uv_buf_t tmpbuf;
-    uv_buf_t *bufs;
+    Py_buffer pbuf;
+    PyObject *callback;
+    uv_buf_t *bufs = NULL;
     uv_udp_send_t *wr = NULL;
     udp_req_data_t *req_data = NULL;
-    udp_send_data_t *send_data = NULL;
-    Py_ssize_t data_len;
-    PyObject *item, *data, *callback;
+    udp_send_data_t *write_data = NULL;
 
     buf_count = 0;
     callback = Py_None;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+
+    if (!PyArg_ParseTuple(args, "(si)s*|O:send", &dest_ip, &dest_port, &pbuf, &callback)) {
         return NULL;
     }
 
-    if (!PyArg_ParseTuple(args, "(si)O|O:send", &dest_ip, &dest_port, &data, &callback)) {
+    if (callback != Py_None && !PyCallable_Check(callback)) {
+        PyBuffer_Release(&pbuf);
+        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
         return NULL;
     }
 
-    if (!PySequence_Check(data) || PyUnicode_Check(data)) {
-        PyErr_SetString(PyExc_TypeError, "only strings and iterables are supported");
+    if (dest_port < 0 || dest_port > 65536) {
+        PyErr_SetString(PyExc_ValueError, "port must be between 0 and 65536");
+        return NULL;
+    }
+
+    if (uv_inet_pton(AF_INET, dest_ip, &addr4) == 1) {
+        address_type = AF_INET;
+    } else if (uv_inet_pton(AF_INET6, dest_ip, &addr6) == 1) {
+        address_type = AF_INET6;
+    } else {
+        PyErr_SetString(PyExc_ValueError, "invalid IP address");
+        return NULL;
+    }
+
+    wr = (uv_udp_send_t *)PyMem_Malloc(sizeof(uv_udp_send_t));
+    if (!wr) {
+        PyErr_NoMemory();
+        goto error;
+    }
+
+    req_data = (udp_req_data_t*) PyMem_Malloc(sizeof(udp_req_data_t));
+    if (!req_data) {
+        PyErr_NoMemory();
+        goto error;
+    }
+
+    write_data = (udp_send_data_t *) PyMem_Malloc(sizeof(udp_send_data_t));
+    if (!write_data) {
+        PyErr_NoMemory();
+        goto error;
+    }
+
+    Py_INCREF(callback);
+    req_data->callback = callback;
+    wr->data = (void *)req_data;
+
+    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t));
+    if (!bufs) {
+        PyErr_NoMemory();
+        goto error;
+    }
+
+    bufs[0] = uv_buf_init(pbuf.buf, pbuf.len);
+    buf_count = 1;
+
+    write_data->bufs = bufs;
+    write_data->buf_count = buf_count;
+    write_data->view = pbuf;
+    req_data->data = write_data;
+
+    if (address_type == AF_INET) {
+        r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+    } else {
+        r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+    }
+    if (r != 0) {
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
+        goto error;
+    }
+
+    Py_RETURN_NONE;
+
+error:
+    PyBuffer_Release(&pbuf);
+    if (bufs) {
+        PyMem_Free(bufs);
+    }
+    if (write_data) {
+        PyMem_Free(write_data);
+    }
+    if (req_data) {
+        Py_DECREF(callback);
+        PyMem_Free(req_data);
+    }
+    if (wr) {
+        PyMem_Free(wr);
+    }
+    return NULL;
+}
+
+
+static PyObject *
+UDP_func_sendlines(UDP *self, PyObject *args)
+{
+    int i, r, buf_count, dest_port, address_type;
+    char *data_str, *tmp, *dest_ip;
+    const char *default_encoding;
+    struct in_addr addr4;
+    struct in6_addr addr6;
+    Py_buffer pbuf;
+    Py_ssize_t data_len, n;
+    PyObject *callback, *seq, *iter, *item, *encoded;
+    uv_buf_t tmpbuf;
+    uv_buf_t *bufs, *new_bufs;
+    uv_udp_send_t *wr = NULL;
+    udp_req_data_t *req_data = NULL;
+    udp_send_data_t *write_data = NULL;
+
+    buf_count = 0;
+    bufs = new_bufs = NULL;
+    callback = Py_None;
+    default_encoding = PyUnicode_GetDefaultEncoding();
+
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
+
+    if (!PyArg_ParseTuple(args, "(si)O|O:sendlines", &dest_ip, &dest_port, &seq, &callback)) {
         return NULL;
     }
 
@@ -334,97 +398,148 @@ UDP_func_send(UDP *self, PyObject *args)
         return NULL;
     }
 
-    wr = (uv_udp_send_t *) PyMem_Malloc(sizeof(uv_udp_send_t));
+    wr = (uv_udp_send_t *)PyMem_Malloc(sizeof(uv_udp_send_t));
     if (!wr) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data = (udp_req_data_t *) PyMem_Malloc(sizeof(udp_req_data_t));
+    req_data = (udp_req_data_t*) PyMem_Malloc(sizeof(udp_req_data_t));
     if (!req_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    send_data = (udp_send_data_t *) PyMem_Malloc(sizeof(udp_send_data_t));
-    if (!send_data) {
+    write_data = (udp_send_data_t *) PyMem_Malloc(sizeof(udp_send_data_t));
+    if (!write_data) {
         PyErr_NoMemory();
         goto error;
     }
 
-    req_data->obj = (PyObject *)self;
     Py_INCREF(callback);
     req_data->callback = callback;
     wr->data = (void *)req_data;
 
-    if (PyString_Check(data)) {
-        /* We have a single string */
-        bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t));
-        if (!bufs) {
-            PyErr_NoMemory();
-            goto error;
-        }
-        data_len = PyString_Size(data);
-        data_str = PyString_AsString(data);
-        tmp = (char *) PyMem_Malloc(data_len + 1);
-        if (!tmp) {
-            PyMem_Free(bufs);
-            PyErr_NoMemory();
-            goto error;
-        }
-        memcpy(tmp, data_str, data_len + 1);
-        tmpbuf = uv_buf_init(tmp, data_len);
-        bufs[0] = tmpbuf;
-        buf_count = 1;
-    } else {
-        /* We have a list */
-        buf_count = 0;
-        n = PySequence_Length(data);
-        bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t) * n);
-        if (!bufs) {
-            PyErr_NoMemory();
-            goto error;
-        }
-        for (i = 0;i < n; i++) {
-            item = PySequence_GetItem(data, i);
-            if (!item || !PyString_Check(item)) {
-                Py_XDECREF(item);
-                continue;
-            }
-            data_len = PyString_Size(item);
-            data_str = PyString_AsString(item);
-            tmp = (char *) PyMem_Malloc(data_len + 1);
-            if (!tmp) {
-                Py_DECREF(item);
-                continue;
-            }
-            memcpy(tmp, data_str, data_len + 1);
-            tmpbuf = uv_buf_init(tmp, data_len);
-            bufs[i] = tmpbuf;
-            buf_count++;
-            Py_DECREF(item);
-        }
-
-    }
-
-    send_data->bufs = bufs;
-    send_data->buf_count = buf_count;
-    req_data->data = (void *)send_data;
-
-    if (address_type == AF_INET) {
-        r = uv_udp_send(wr, self->uv_handle, bufs, buf_count, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
-    } else {
-        r = uv_udp_send6(wr, self->uv_handle, bufs, buf_count, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
-    }
-    if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+    iter = PyObject_GetIter(seq);
+    if (iter == NULL) {
         goto error;
     }
+
+    n = iter_guess_size(iter, 8);   /* if we can't get the size hint, preallocate 8 slots */
+    bufs = (uv_buf_t *) PyMem_Malloc(sizeof(uv_buf_t) * n);
+    if (!bufs) {
+        Py_DECREF(iter);
+        PyErr_NoMemory();
+        goto error;
+    }
+
+    i = 0;
+    while (1) {
+        item = PyIter_Next(iter);
+        if (item == NULL) {
+            if (PyErr_Occurred()) {
+                Py_DECREF(iter);
+                goto error;
+            } else {
+                /* StopIteration */
+                break;
+            }
+        }
+
+        if (PyUnicode_Check(item)) {
+            encoded = PyUnicode_AsEncodedString(item, default_encoding, "strict");
+            if (encoded == NULL) {
+                Py_DECREF(item);
+                Py_DECREF(iter);
+                goto error;
+            }
+            data_str = PyString_AS_STRING(encoded);
+            data_len = PyString_GET_SIZE(encoded);
+            tmp = (char *) PyMem_Malloc(data_len);
+            if (!tmp) {
+                Py_DECREF(item);
+                Py_DECREF(iter);
+                PyErr_NoMemory();
+                goto error;
+            }
+            memcpy(tmp, data_str, data_len);
+            tmpbuf = uv_buf_init(tmp, data_len);
+        } else {
+            if (PyObject_GetBuffer(item, &pbuf, PyBUF_CONTIG_RO) < 0) {
+                Py_DECREF(item);
+                Py_DECREF(iter);
+                goto error;
+            }
+            data_str = pbuf.buf;
+            data_len = pbuf.len;
+            tmp = (char *) PyMem_Malloc(data_len);
+            if (!tmp) {
+                Py_DECREF(item);
+                Py_DECREF(iter);
+                PyBuffer_Release(&pbuf);
+                PyErr_NoMemory();
+                goto error;
+            }
+            memcpy(tmp, data_str, data_len);
+            tmpbuf = uv_buf_init(tmp, data_len);
+            PyBuffer_Release(&pbuf);
+        }
+
+        /* Check if we allocated enough space */
+        if (buf_count+1 < n) {
+            /* we have enough size */
+        } else {
+            /* preallocate 8 more slots */
+            n += 8;
+            new_bufs = (uv_buf_t *) PyMem_Realloc(bufs, sizeof(uv_buf_t) * n);
+            if (!new_bufs) {
+                Py_DECREF(item);
+                Py_DECREF(iter);
+                PyErr_NoMemory();
+                goto error;
+            }
+            bufs = new_bufs;
+        }
+        bufs[i] = tmpbuf;
+        i++;
+        buf_count++;
+        Py_DECREF(item);
+    }
+    Py_DECREF(iter);
+
+    /* we may have over allocated space, shrink it to the minimum required */
+    new_bufs = (uv_buf_t *) PyMem_Realloc(bufs, sizeof(uv_buf_t) * buf_count);
+    if (!new_bufs) {
+        PyErr_NoMemory();
+        goto error;
+    }
+    bufs = new_bufs;
+
+    write_data->bufs = bufs;
+    write_data->buf_count = buf_count;
+    req_data->data = write_data;
+
+    if (address_type == AF_INET) {
+        r = uv_udp_send(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip4_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+    } else {
+        r = uv_udp_send6(wr, (uv_udp_t *)UV_HANDLE(self), bufs, buf_count, uv_ip6_addr(dest_ip, dest_port), (uv_udp_send_cb)on_udp_send);
+    }
+    if (r != 0) {
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
+        goto error;
+    }
+
     Py_RETURN_NONE;
 
 error:
-    if (send_data) {
-        PyMem_Free(send_data);
+    if (bufs) {
+        for (i = 0; i < buf_count; i++) {
+            PyMem_Free(bufs[i].base);
+        }
+        PyMem_Free(bufs);
+    }
+    if (write_data) {
+        PyMem_Free(write_data);
     }
     if (req_data) {
         Py_DECREF(callback);
@@ -445,18 +560,15 @@ UDP_func_set_membership(UDP *self, PyObject *args)
 
     interface_address = NULL;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "si|s:set_membership", &multicast_address, &membership, &interface_address)) {
         return NULL;
     }
 
-    r = uv_udp_set_membership(self->uv_handle, multicast_address, interface_address, membership);
+    r = uv_udp_set_membership((uv_udp_t *)UV_HANDLE(self), multicast_address, interface_address, membership);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -464,38 +576,6 @@ UDP_func_set_membership(UDP *self, PyObject *args)
 }
 
 
-static PyObject *
-UDP_func_close(UDP *self, PyObject *args)
-{
-    PyObject *callback = Py_None;
-
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "UDP is already closed");
-        return NULL;
-    }
-
-    if (!PyArg_ParseTuple(args, "|O:close", &callback)) {
-        return NULL;
-    }
-
-    if (callback != Py_None && !PyCallable_Check(callback)) {
-        PyErr_SetString(PyExc_TypeError, "a callable or None is required");
-        return NULL;
-    }
-
-    Py_INCREF(callback);
-    self->on_close_cb = callback;
-
-    /* Increase refcount so that object is not removed before the callback is called */
-    Py_INCREF(self);
-
-    uv_close((uv_handle_t *)self->uv_handle, on_udp_close);
-    self->uv_handle = NULL;
-
-    Py_RETURN_NONE;
-}
-
-
 static PyObject *
 UDP_func_getsockname(UDP *self)
 {
@@ -507,14 +587,11 @@ UDP_func_getsockname(UDP *self)
 
     namelen = sizeof(sockname);
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
-    r = uv_udp_getsockname(self->uv_handle, &sockname, &namelen);
+    r = uv_udp_getsockname((uv_udp_t *)UV_HANDLE(self), &sockname, &namelen);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -538,10 +615,7 @@ UDP_func_set_multicast_ttl(UDP *self, PyObject *args)
 {
     int r, ttl;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:set_multicast_ttl", &ttl)) {
         return NULL;
@@ -552,9 +626,9 @@ UDP_func_set_multicast_ttl(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_multicast_ttl(self->uv_handle, ttl);
+    r = uv_udp_set_multicast_ttl((uv_udp_t *)UV_HANDLE(self), ttl);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -568,18 +642,15 @@ UDP_func_set_broadcast(UDP *self, PyObject *args)
     int r;
     PyObject *enable;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "already closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!:set_broadcast", &PyBool_Type, &enable)) {
         return NULL;
     }
 
-    r = uv_udp_set_broadcast(self->uv_handle, (enable == Py_True) ? 1 : 0);
+    r = uv_udp_set_broadcast((uv_udp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -593,18 +664,15 @@ UDP_func_set_multicast_loop(UDP *self, PyObject *args)
     int r;
     PyObject *enable;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "already closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "O!:set_multicast_loop", &PyBool_Type, &enable)) {
         return NULL;
     }
 
-    r = uv_udp_set_multicast_loop(self->uv_handle, (enable == Py_True) ? 1 : 0);
+    r = uv_udp_set_multicast_loop((uv_udp_t *)UV_HANDLE(self), (enable == Py_True) ? 1 : 0);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -617,10 +685,7 @@ UDP_func_set_ttl(UDP *self, PyObject *args)
 {
     int r, ttl;
 
-    if (!self->uv_handle) {
-        PyErr_SetString(PyExc_UDPError, "closed");
-        return NULL;
-    }
+    RAISE_IF_HANDLE_CLOSED(self, PyExc_HandleClosedError, NULL);
 
     if (!PyArg_ParseTuple(args, "i:set_multicast_ttl", &ttl)) {
         return NULL;
@@ -631,9 +696,9 @@ UDP_func_set_ttl(UDP *self, PyObject *args)
         return NULL;
     }
 
-    r = uv_udp_set_ttl(self->uv_handle, ttl);
+    r = uv_udp_set_ttl((uv_udp_t *)UV_HANDLE(self), ttl);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         return NULL;
     }
 
@@ -651,7 +716,7 @@ UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
 
     UNUSED_ARG(kwargs);
 
-    if (self->uv_handle) {
+    if (UV_HANDLE(self)) {
         PyErr_SetString(PyExc_UDPError, "Object already initialized");
         return -1;
     }
@@ -660,9 +725,9 @@ UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
         return -1;
     }
 
-    tmp = (PyObject *)self->loop;
+    tmp = (PyObject *)((Handle *)self)->loop;
     Py_INCREF(loop);
-    self->loop = loop;
+    ((Handle *)self)->loop = loop;
     Py_XDECREF(tmp);
 
     uv_udp_handle = PyMem_Malloc(sizeof(uv_udp_t));
@@ -671,14 +736,14 @@ UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
         Py_DECREF(loop);
         return -1;
     }
-    r = uv_udp_init(UV_LOOP(self), uv_udp_handle);
+    r = uv_udp_init(UV_HANDLE_LOOP(self), uv_udp_handle);
     if (r != 0) {
-        raise_uv_exception(self->loop, PyExc_UDPError);
+        RAISE_UV_EXCEPTION(UV_HANDLE_LOOP(self), PyExc_UDPError);
         Py_DECREF(loop);
         return -1;
     }
     uv_udp_handle->data = (void *)self;
-    self->uv_handle = uv_udp_handle;
+    UV_HANDLE(self) = (uv_handle_t *)uv_udp_handle;
 
     return 0;
 }
@@ -687,11 +752,10 @@ UDP_tp_init(UDP *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 UDP_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 {
-    UDP *self = (UDP *)PyType_GenericNew(type, args, kwargs);
+    UDP *self = (UDP *)HandleType.tp_new(type, args, kwargs);
     if (!self) {
         return NULL;
     }
-    self->uv_handle = NULL;
     return (PyObject *)self;
 }
 
@@ -699,9 +763,8 @@ UDP_tp_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 static int
 UDP_tp_traverse(UDP *self, visitproc visit, void *arg)
 {
-    Py_VISIT(self->data);
     Py_VISIT(self->on_read_cb);
-    Py_VISIT(self->loop);
+    HandleType.tp_traverse((PyObject *)self, visit, arg);
     return 0;
 }
 
@@ -709,32 +772,19 @@ UDP_tp_traverse(UDP *self, visitproc visit, void *arg)
 static int
 UDP_tp_clear(UDP *self)
 {
-    Py_CLEAR(self->data);
     Py_CLEAR(self->on_read_cb);
-    Py_CLEAR(self->loop);
+    HandleType.tp_clear((PyObject *)self);
     return 0;
 }
 
 
-static void
-UDP_tp_dealloc(UDP *self)
-{
-    if (self->uv_handle) {
-        uv_close((uv_handle_t *)self->uv_handle, on_udp_dealloc_close);
-        self->uv_handle = NULL;
-    }
-    UDP_tp_clear(self);
-    Py_TYPE(self)->tp_free((PyObject *)self);
-}
-
-
 static PyMethodDef
 UDP_tp_methods[] = {
     { "bind", (PyCFunction)UDP_func_bind, METH_VARARGS, "Bind to the specified IP and port." },
     { "start_recv", (PyCFunction)UDP_func_start_recv, METH_VARARGS, "Start accepting data." },
     { "stop_recv", (PyCFunction)UDP_func_stop_recv, METH_NOARGS, "Stop receiving data." },
     { "send", (PyCFunction)UDP_func_send, METH_VARARGS, "Send data over UDP." },
-    { "close", (PyCFunction)UDP_func_close, METH_VARARGS, "Close UDP connection." },
+    { "sendlines", (PyCFunction)UDP_func_sendlines, METH_VARARGS, "Send a sequence of data over UDP." },
     { "getsockname", (PyCFunction)UDP_func_getsockname, METH_NOARGS, "Get local socket information." },
     { "set_membership", (PyCFunction)UDP_func_set_membership, METH_VARARGS, "Set membership for multicast address." },
     { "set_multicast_ttl", (PyCFunction)UDP_func_set_multicast_ttl, METH_VARARGS, "Set the multicast TTL." },
@@ -745,19 +795,12 @@ UDP_tp_methods[] = {
 };
 
 
-static PyMemberDef UDP_tp_members[] = {
-    {"loop", T_OBJECT_EX, offsetof(UDP, loop), READONLY, "Loop where this UDP is running on."},
-    {"data", T_OBJECT, offsetof(UDP, data), 0, "Arbitrary data."},
-    {NULL}
-};
-
-
 static PyTypeObject UDPType = {
     PyVarObject_HEAD_INIT(NULL, 0)
     "pyuv.UDP",                                                     /*tp_name*/
     sizeof(UDP),                                                    /*tp_basicsize*/
     0,                                                              /*tp_itemsize*/
-    (destructor)UDP_tp_dealloc,                                     /*tp_dealloc*/
+    0,                                                              /*tp_dealloc*/
     0,                                                              /*tp_print*/
     0,                                                              /*tp_getattr*/
     0,                                                              /*tp_setattr*/
@@ -781,7 +824,7 @@ static PyTypeObject UDPType = {
     0,                                                              /*tp_iter*/
     0,                                                              /*tp_iternext*/
     UDP_tp_methods,                                                 /*tp_methods*/
-    UDP_tp_members,                                                 /*tp_members*/
+    0,                                                              /*tp_members*/
     0,                                                              /*tp_getsets*/
     0,                                                              /*tp_base*/
     0,                                                              /*tp_dict*/
