@@ -1,4 +1,5 @@
 
+import functools
 import threading
 import time
 
@@ -18,14 +19,18 @@ class ThreadPoolTest(unittest2.TestCase):
         self.pool_cb_called += 1
         time.sleep(0.1)
 
-    def after_work_cb(self, status, result):
-        self.assertEqual(status, 0)
+    def after_work_cb(self, result, error):
+        self.assertEqual(error, None)
         self.pool_after_work_cb_called += 1
 
     def test_threadpool1(self):
         self.pool.queue_work(self.run_in_pool, self.after_work_cb)
-        self.pool.queue_work(self.run_in_pool, self.after_work_cb, (1, 2, 3), {})
-        self.pool.queue_work(self.run_in_pool, self.after_work_cb, (), {'test': 1})
+        args = (1, 2, 3)
+        kw = {}
+        self.pool.queue_work(functools.partial(self.run_in_pool, *args, **kw), self.after_work_cb)
+        args = ()
+        kw = {'test': 1}
+        self.pool.queue_work(functools.partial(self.run_in_pool, *args, **kw), self.after_work_cb)
         self.loop.run()
         self.assertEqual(self.pool_cb_called, 3)
         self.assertEqual(self.pool_after_work_cb_called, 3)
@@ -33,8 +38,8 @@ class ThreadPoolTest(unittest2.TestCase):
     def raise_in_pool(self, *args, **kw):
         1/0
 
-    def after_work_cb2(self, status, result):
-        self.assertEqual(status, -1)
+    def after_work_cb2(self, result, error):
+        self.assertEqual(error[0], ZeroDivisionError)
 
     def test_threadpool2(self):
         self.pool.queue_work(self.raise_in_pool, self.after_work_cb2)
@@ -78,9 +83,7 @@ class ThreadPoolThreadCount(unittest2.TestCase):
         time.sleep(0.1)
 
     def test_threadpool_thread_count(self):
-        nthreads = self.pool.get_nthreads()
-        self.pool.set_min_parallel_threads(10)
-        self.pool.set_max_parallel_threads(10)
+        self.pool.set_parallel_threads(10)
         self.pool.queue_work(self.run_in_pool)
         self.pool.queue_work(self.run_in_pool)
         self.pool.queue_work(self.run_in_pool)
@@ -91,8 +94,6 @@ class ThreadPoolThreadCount(unittest2.TestCase):
         self.pool.queue_work(self.run_in_pool)
         self.pool.queue_work(self.run_in_pool)
         self.loop.run()
-        after_nthreads = self.pool.get_nthreads()
-        self.assertTrue(after_nthreads > nthreads)
 
 
 if __name__ == '__main__':
